-- Freeslots
freeslot("SPR_BEBZ")

freeslot("MT_BEEBZVISUAL")

freeslot("S_BEEBZVISUAL")

freeslot("sfx_bebzc1", "sfx_bebzc2", "sfx_bebzc3", "sfx_bebzc4")

-- ===========================================================================
-- = Configuration goes here
-- ===========================================================================

local skin_name = "beebz" -- Skin id

local ability_timer = 2*TICRATE -- Cooldown for ability

-- Scale for visual popup

-- Initial scale
local showability_scale = FRACUNIT/50

-- Final scale
local showability_destscale = FRACUNIT/5

-- Time to scale from initial scale to final scale
local showability_growtime = 10

-- Base offset from player z pposition
local showability_base_offset = 40*FRACUNIT

-- Maximum extra offset, for animation
local showability_offset = 30*FRACUNIT

local abilities = {
    -- No ability active
    [0] = {
        speedmod = 0, -- Speed increase
        weightmod = 0, -- Weight increase
        sfx = sfx_bebzc4, -- Sound effect
		frame = D, -- Visual for popup
    },

    [1] = {
        speedmod = 2,
        weightmod = 0,
        sfx = sfx_bebzc1,
		frame = A,
    },

    -- Etc
    [2] = {
        speedmod = 0,
        weightmod = 2,
        sfx = sfx_bebzc2,
		frame = B,
    },

    [3] = {
        speedmod = 0,
        weightmod = -2,
        sfx = sfx_bebzc3,
		frame = C,
    },
}

-- Cvars

-- Allow using Beebz ability
local cv_beebz = CV_RegisterVar({
    name = "beebz_allow_ability",
    defaultvalue = "On",
    possiblevalue = CV_OnOff,
    flags = CV_NETVAR,
})

-- Allow using ability in time attack
local cv_beebz_ta = CV_RegisterVar({
    name = "beebz_allow_ability_ta",
    defaultvalue = "Off",
    possiblevalue = CV_OnOff,
    flags = CV_NETVAR,
})

-- Allow using ability when under effect of restat
local cv_beebz_restat = CV_RegisterVar({
    name = "beebz_allow_ability_restat",
    defaultvalue = "On",
    possiblevalue = CV_OnOff,
    flags = CV_NETVAR,
})

-- ===========================================================================
-- = End of configuration
-- ===========================================================================

mobjinfo[MT_BEEBZVISUAL] = {
	-- MF_MISSILE has a unique property that no other type of object has:
	-- if the mobj's target is a player and the player respawns,
	-- the mobj's target is automatically updated to the player's new mobj
	-- (ab)use this to keep the visual across respawns
	flags = MF_MISSILE|MF_SCENERY|MF_NOGRAVITY|MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_DONTENCOREMAP,
	spawnstate = S_BEEBZVISUAL,
}

states[S_BEEBZVISUAL] = {
    sprite = SPR_BEBZ,
}

local function countPlayers()
    local c = 0

    for player in players.iterate do
        if not player.spectator then c = c + 1 end
    end

    return c
end

local function clamp(minval, val, maxval) return max(minval, min(maxval, val)) end

-- thinker for visual
addHook("MobjThinker", function(mo)
	local pmo = mo.target
    local t = min(FixedDiv((ability_timer-mo.fuse)*FRACUNIT, ability_timer/2*FRACUNIT), FRACUNIT)

	if not (pmo and pmo.player.spectator == false and pmo.skin == skin_name) then
		mo.fuse = 1 -- don't need this visual anymore
		return
	end

    t = FixedSqrt(t) -- For more smooth animation

    P_MoveOrigin(mo, pmo.x, pmo.y, pmo.z + showability_base_offset + FixedMul(t, FixedMul(showability_offset, mapobjectscale)))
end, MT_BEEBZVISUAL)

-- Executed once ability is activated
local function applyAbility(num, pmo, beebz)
	-- First tap activates ability, second tap disables it
	if num == beebz.ability then num = 0 end

	beebz.ability = num
	beebz.abilitytimer = ability_timer

    beebz.speedmod = abilities[num].speedmod or 0
    beebz.weightmod = abilities[num].weightmod or 0

	beebz.visual = P_SpawnMobj(pmo.x, pmo.y, pmo.z + showability_base_offset, MT_BEEBZVISUAL)
	local vis = beebz.visual
	vis.frame = abilities[num].frame
	vis.target = pmo
	vis.fuse = ability_timer
    vis.scale = FixedMul(showability_scale, mapobjectscale)
    vis.destscale = FixedMul(showability_destscale, mapobjectscale)
    vis.scalespeed = (vis.destscale - vis.scale)/showability_growtime

    S_StartSound(pmo, abilities[num].sfx)
end

-- Commands to alternative ability activation
local function commandAbility(num)
    return function(player)
        local beebz = player.beebz

        if not (beebz and beebz.abilitytimer == 0 and player.mo) then return end

        applyAbility(num, player.mo, beebz)
    end
end

COM_AddCommand("beebz_ability1", commandAbility(1))
COM_AddCommand("beebz_ability2", commandAbility(2))
COM_AddCommand("beebz_ability3", commandAbility(3))

-- Clean up beebz states if needed
local function beebzCleanup(player)
	local skin = player.mo.skin

	player.kartspeed = player.HMRs or skins[skin].kartspeed
	player.kartweight = player.HMRw or skins[skin].kartweight

	if player.beebz.visual and player.beebz.visual.valid then P_RemoveMobj(player.beebz.visual) end
	player.beebz = nil
end

-- Disable BT_CUSTOM* input in combi
local cv_combi

local skindata

addHook("ThinkFrame", function()
    local numplayers = countPlayers()

	if leveltime == 0 then return end

    for player in players.iterate do
	    if not player.mo then continue end
		local pmo = player.mo

		-- in order of decreasing priority...
		if pmo.skin ~= skin_name
		or (numplayers <= 1 and (not cv_beebz_ta.value)
		or (player.HMRs and (not cv_beebz_restat.value)
		or (not cv_beebz.value))) then
			-- nuh-uh
			if player.beebz then beebzCleanup(player) end
			continue
		else
			-- yup
			if not player.beebz then
				player.beebz = {
					speedmod = 0,
					weightmod = 0,
					ability = 0,
					abilitytimer = 0,
					c1 = 0,
					c2 = 0,
					c3 = 0,
				}
			end
		end

		local beebz = player.beebz

		local combi = cv_combi and cv_combi.value

		-- First, lets have nice convinient input data (thanks hornmod for idea of
		-- how to organize this)
		local buttons = 0

		-- If combi is active, don't use buttons. Bind commands for that instead
		if not combi then
			buttons = player.cmd.buttons
		end

		beebz.c1 = (buttons & BT_CUSTOM1) and $ + 1 or 0
		beebz.c2 = (buttons & BT_CUSTOM2) and $ + 1 or 0
		beebz.c3 = (buttons & BT_CUSTOM3) and $ + 1 or 0

		-- If we can use ability, check for input
		if beebz.abilitytimer == 0 then
			local newability
			if beebz.c1 == 1 then
				newability = 1
			elseif beebz.c2 == 1 then
				newability = 2
			elseif beebz.c3 == 1 then
				newability = 3
			end

			if newability then applyAbility(newability, pmo, beebz) end
		else
			beebz.abilitytimer = $ - 1
		end

		local speed = player.HMRs or skindata.kartspeed
		local weight = player.HMRw or skindata.kartweight

		-- No funni 255 speed/weight
		player.kartspeed = clamp(1, speed + beebz.speedmod, 9)
		player.kartweight = clamp(1, weight + beebz.weightmod, 9)
    end
end)

local function dataLookup()
    if not skindata then skindata = skins[skin_name] end -- Cache skin data
	if not cv_combi then cv_combi = CV_FindVar("combi_active") end
end

addHook("MapLoad", function()
	dataLookup()
    for player in players.iterate do
        if player.beebz then beebzCleanup(player) end
    end
end)

-- MapLoad isn't called for mid-joiners
addHook("NetVars", dataLookup)