local doconfig = false

local cv_keepcolor = CV_RegisterVar({
	name = "ditto_keepcolor",
	defaultvalue = "Off",
	flags = CV_CALL|CV_NOINIT,
	possiblevalue = CV_OnOff,
	func = do doconfig = true end
})

local cv_nostartrandom = CV_RegisterVar({
	name = "ditto_nostartrandom",
	defaultvalue = "Off",
	flags = CV_CALL|CV_NOINIT,
	possiblevalue = CV_OnOff,
	func = do doconfig = true end
})

local cv_transformonce = CV_RegisterVar({
	name = "ditto_transformonce",
	defaultvalue = "Off",
	flags = CV_CALL|CV_NOINIT,
	possiblevalue = CV_OnOff,
	func = do doconfig = true end
})

local KEEP_COLOR = 1
local DONT_START_RANDOM = 2
local TRANSFORM_ONLY_ONCE = 4

local function sendconfig()
	local c = (cv_keepcolor.value * KEEP_COLOR)
	+ (cv_nostartrandom.value * DONT_START_RANDOM)
	+ (cv_transformonce.value * TRANSFORM_ONLY_ONCE)

	--print("Sending "..c)
	COM_BufAddText(consoleplayer, "dittoconfig "..string.char(c + 65))
end

COM_AddCommand("dittoconfig", function(p, c)
	--print("From "..p.name.." comes "..tostring(c))
	c = tonumber(c or "", 18) -- kek
	if c and c >= 10 and c <= 17 then p.dittoconfig = c - 10 end
end)

-- config over... phew...

freeslot("S_DITTO_PARTICLE", "sfx_incopy", "sfx_decopy")

-- the fx stuff, make sure it looks and sticks well
local function dittoParticle(mo)
	local target = mo.target
	if target and target.valid then
		P_MoveOrigin(mo, target.x, target.y, target.z)
		mo.frame = 5 - (mo.fuse-1) / 3
	else
		mo.fuse = 1 -- die
	end
end

states[S_DITTO_PARTICLE] = {
	sprite = SPR_CLAS,
	tics = 1,
	action = dittoParticle,
	nextstate = S_DITTO_PARTICLE,
}

local MAXPLAYERS = #players

--woaaaaa
local DITTO_SKIN = 'ditto'

-- we all know this one... right?
local GO_TIME = 6 * TICRATE + (3 * TICRATE / 4)

-- all the arbitrary gameplay parameters I decided.
local BASE_TRANSFORM_TIME = 15*TICRATE / MAXPLAYERS
local MIN_TRANSFORM_TIME = 7*TICRATE
local TRANFORMING_DURATION = 3*TICRATE / MAXPLAYERS

local START_TRANSFORM_TIMEMARK = 4*TICRATE
local ENDRACE_DETRANSFORM_TIME = 2*TICRATE / MAXPLAYERS

local RETAIN_MEMORY_DISTANCE = 2000*FRACUNIT
local COPY_ENABLED_DISTANCE = 1000*FRACUNIT

-- this calls the transforming effects on a given mobj.
local function _copyFX(pmo, sfx)
	local qsfx = (sfx == sfx_incopy and sfx_decopy or sfx_incopy)

	local copy_particle = P_SpawnMobj(pmo.x, pmo.y, pmo.z, MT_THOK)
	copy_particle.target = pmo
	copy_particle.fuse = 18
	copy_particle.state = S_DITTO_PARTICLE

	S_StopSoundByID(pmo, qsfx)
	S_StartSound(pmo, sfx)
end

-- yeah, it kind of bothers me that we have to go through the players iteration loop
-- everytime we need the number of players, so I figured we at least could only do it
-- only once a frame (although now that I think about it, it's not even usefull in this
-- script… SHIT!)
local function countNonSpecPlayers()
    local n = 0
    for p in players.iterate do
		if not p.spectator then n = $+1 end
    end
    return n
end

-- tells if a player's... yeah the function name says it all...
local function finished(p)
	return not p.spectator and ((p.pflags & PF_TIMEOVER) or p.exiting) 
end

-- tells if a particular targeted player is copyable by a ditto (that's the character I'm
-- making here, remember?)
local function canCopy(target)
	return not (target.spectator or target.exiting or target.kartstuff[k_hyudorotimer])
	and target.playerstate ~= PST_DEAD and target.mo.skin ~= DITTO_SKIN
	and (G_RaceGametype() or target.kartstuff[k_bumper] > 0)
end

-- the function that transforms a ditto player into another.
-- if boolean parameter 'fx' is not given (or if 'fx==true')
-- then the transformation effects and sound will be called upon
-- the transfroming players, otherwise (if 'fx==false'), player will
-- transform quietly
local function playerCopies(p, target)
	_copyFX(p.mo, sfx_incopy)

	-- moon crystal power! make u-... wait no, wrong character...
	p.mo.skin = target.mo.skin

	-- anyway, this where we make all the changes and save all the relevant settings and whatnot
	p.dittostuff = $ or {}
	local ds = p.dittostuff
	if not ds.istransformed then
		ds.original_color = p.skincolor
	end

	ds.transform_time = BASE_TRANSFORM_TIME
	ds.copied_target = target
	ds.copied_skin = target.mo.skin
	ds.copied_color = (FRIENDMOD_Active or (p.dittoconfig & KEEP_COLOR))and p.skincolor or target.skincolor
	ds.copying_target = nil
	ds.transform_progress = 0
	ds.real_transform_time = 0

	p.skincolor = ds.copied_color
	p.mo.color = ds.copied_color

	-- save our restat
	if not ds.istransformed and p.HMRs then
		ds.restat_s = p.HMRs
		ds.restat_w = p.HMRw
	end

	-- gotta deal with hostmod's restat
	-- which simple presence locks the character stats
	if target.HMRs then
		p.HMRs = target.HMRs
		p.HMRw = target.HMRw
		p.kartspeed, p.kartweight = p.HMRs, p.HMRw
	else
		p.kartspeed = target.kartspeed
		p.kartweight = target.kartweight
		p.HMRs, p.HMRw = nil, nil
	end

	-- hornmod
	if HORNMOD then
		local dhm = p.hornmod
		-- save original horn
		if not ds.istransformed and dhm and dhm.horn then
			ds.horn = dhm.horn
		end
		if dhm and target.hornmod.horn then
			dhm.horn = target.hornmod.horn
			if dhm.origin then
				S_StopSound(dhm.origin)
			end
		end
	end

	--[[ emotes
	if EMOJI_TIME then
		if not ds.istransformed and p.emoteCustomTable then
			ds.emotes = p.emoteCustomTable
		end

		if target.emoteCustomTable then
			p.emoteCustomTable = target.emoteCustomTable
		end
	end --]]

	-- secondcolor
	if target.ksc_secondcolor ~= nil then
		if FRIENDMOD_Active then
			p.ksc_secondcolor = ds.original_color
		else
			p.ksc_secondcolor = target.ksc_secondcolor
		end
	end

	ds.istransformed = true
end

-- go back! go back!
-- the transformed ditto goes back to it form squishy self
-- (again with 'fx' or not)
local function playerDetransforms(p, fx)
	local ds = p.dittostuff
	local oldspec = p.spectator
	p.spectator = false

	if fx and not oldspec then _copyFX(p.mo, sfx_decopy) end

	ds.transform_time = 0

	if ds.istransformed then
		p.skincolor = ds.original_color
		p.mo.color = ds.original_color
	end

	-- so, if restat is active we might have tampered with the
	--  restat's stats... should put them back
	if ds.restat_s then
		p.HMRs, p.kartspeed = ds.restat_s, ds.restat_s
		p.HMRw, p.kartweight = ds.restat_w, ds.restat_w
		ds.restat_s, ds.restat_w = nil, nil
	else
		p.kartspeed = skins[DITTO_SKIN].kartspeed
		p.kartweight = skins[DITTO_SKIN].kartweight
		p.HMRs, p.HMRw = nil, nil
	end

	-- horn stuff
	if HORNMOD and ds.horn and p.hornmod then
		p.hornmod.horn = ds.horn
		if p.hornmod.origin then
			S_StopSound(p.hornmod.origin)
		end
	end

	--[[ emotes
	if EMOJI_TIME and ds.emotes then
		p.emoteCustomTable = ds.emotes
	end --]]

	-- secondcolor
	p.ksc_secondcolor = nil

	p.mo.skin = DITTO_SKIN
	p.dittostuff = {
		-- keep these, just in case a copy was in progress when ditto detransformed
		transform_progress = ds.transform_progress,
		copying_target = ds.copying_target
	}

	ds.istransformed = false
	p.spectator = oldspec
end

-- I'm paranoid, this a litte system to make sure relevant stuff
--  will be flushed out between various state (race, intermission, vote, etc.)
local _cleaning = 0
local function cleaning(lvl)
	if _cleaning == lvl then return end

	for player in players.iterate do
		if player.dittostuff then
			playerDetransforms(player, false)
		end

		player.dittostuff = nil
	end
	_cleaning = lvl
end

-- make sure we have a fresh start
addHook("MapLoad", function()
	for p in players.iterate do
		if p.dittostuff then
			playerDetransforms(p, false)
			p.dittostuff = nil
		end

		-- trying to make things work nicely with HOSTMOD's restat
		-- (hope I'm not making things worse though, lol)
		if p.HMRs and p.HMRs_p ~= p.HMRs or p.HMRw_p ~= p.HMRw then
			p.HMRs = p.HMRs_p
			p.HMRw = p.HMRw_p
		end
	end
end)

-- and here it is, the 'thinkFrame' hook!
-- where all the rage is at...
addHook("ThinkFrame", function()
	if doconfig then
		sendconfig()
		doconfig = false
	end

	-- we pick up all the candidates for the 'copy random player at start' bit
	local copyCandidates
	if leveltime == START_TRANSFORM_TIMEMARK then
		copyCandidates = {}
		for player in players.iterate do
			if canCopy(player) then
				table.insert(copyCandidates, player)
			end
		end
	end

	-- here we go, the main player loop, everyone should get their due
	for player in players.iterate do
		if leveltime < START_TRANSFORM_TIMEMARK then continue end
		if player.spectator then
			if player.dittostuff then
				playerDetransforms(player, false)
				player.dittostuff = nil
			end
			continue
		end

		local pmo_cur_skin = player.mo.skin

		-- dittos get ditto stuff, man
		if pmo_cur_skin == DITTO_SKIN and not finished(player) and not player.dittostuff then
			player.dittostuff = {}
			player.dittostuff.transform_progress = 0
		end

		local ds = player.dittostuff

		-- sorry, this hook be dittos only
		if not ds then continue end

		if not player.dittoconfig then player.dittoconfig = 0 end

		-- so you're a transformed ditto, huh?
		-- yeah, please don't switch back to the ditto skin nor change you color
		--  manually... this isn't how it's supposed to work bro
		if ds.istransformed then
			if pmo_cur_skin == DITTO_SKIN then
				player.mo.skin = ds.copied_skin
			end

			if player.skincolor ~= ds.copied_color then
				ds.original_color = player.skincolor
				player.mo.color = ds.copied_color
				player.skincolor = ds.copied_color
			end
		end

		-- oh, so you're changing to a completely unrelated skin?
		-- yeah get out of here!
		if pmo_cur_skin ~= DITTO_SKIN and pmo_cur_skin ~= ds.copied_skin then
			playerDetransforms(player, false)
			player.mo.skin = pmo_cur_skin
			player.dittostuff = nil
			continue
		end

		-- let's go! if the 'no_random' setting isn't on, random copy stuff
		if leveltime == START_TRANSFORM_TIMEMARK and not (player.dittoconfig & DONT_START_RANDOM) then
			local target, pcount
			if not copyCandidates or #copyCandidates <= 0 then
				pcount = countNonSpecPlayers()
				if pcount < 2 then return end
				local random = P_RandomRange(0, pcount-2)
				local j = 0
				for p in players.iterate do
					if p ~= player and not p.spectator and canCopy(p) then
						target = p
						if j == random then break end
						j = min(j + 1, random)
					end
				end
			else
				pcount = #copyCandidates
				local random = P_RandomRange(0, pcount-1)
				for i = 0, pcount-1 do
					local p = copyCandidates[((random+i)%pcount)+1]
					if canCopy(p) then target = p end
				end
			end

			if target then playerCopies(player, target) end
			ds.transform_locked = (player.dittoconfig & TRANSFORM_ONLY_ONCE) > 0
		end

		-- you dead bro? well what can I do, heh?
		if player.playerstate == PST_DEAD then continue end

		-- been transormet for how long, bro?
		ds.real_transform_time = ds.istransformed and $ + 1 or 0

		if leveltime % MAXPLAYERS == #player then
			-- so you're all done?  well de-transform, play dumb sounds or whatever
			if finished(player) then
				if ds.istransformed then
					ds.transform_time = max(0, min($ - 1, ENDRACE_DETRANSFORM_TIME))
					if ds.transform_time == 0 then
						S_StopSoundByID(player.mo, skins[pmo_cur_skin].soundsid[SKSKWIN])
						S_StopSoundByID(player.mo, skins[pmo_cur_skin].soundsid[SKSKLOSE])
						playerDetransforms(player, true)
						player.dittostuff = nil
						S_StartSound(player ~= consoleplayer and player.mo or nil, skins[DITTO_SKIN].soundsid[K_IsPlayerLosing(player) and SKSKLOSE or SKSKWIN])
					end
				end
			-- shit is on. Now, who to copy? hmmm...
			elseif leveltime > GO_TIME and not ds.transform_locked then
				-- invalidate targets
				if ds.copied_target and (not ds.copied_target.valid or ds.copied_target.spectator) then
					ds.copied_target = nil
				end
				if ds.copying_target and (not ds.copying_target.valid or ds.copying_target.spectator) then
					ds.copying_target = nil
				end

				local closest_p
				local cur_target_distance = INT32_MAX

				-- yeah, so we make one quick loop through the players and try to find the nearest target,
				-- and compue all the relevant distances
				local min_distance = INT32_MAX
				local race = G_RaceGametype()

				for other_p in players.iterate do
					if other_p ~= player and not (other_p.spectator or other_p.kartstuff[k_hyudorotimer])
					and other_p.playerstate ~= PST_DEAD and other_p.mo.skin ~= DITTO_SKIN
					and (race or other_p.kartstuff[k_bumper] > 0) then
						local d = FixedDiv(R_PointToDist2(player.mo.x, player.mo.y, other_p.mo.x, other_p.mo.y), mapobjectscale)
						if ds.copied_target and other_p.mo.skin == ds.copied_skin then
							cur_target_distance = d
						end
						if d < min_distance and d <= COPY_ENABLED_DISTANCE then
							min_distance = d
							closest_p = other_p
						end
					end
				end

				ds.untransforming = false

				-- if the guy we copied is still in range and/or the closest
				-- ('closest' also means he's necessarily in copy range, mind you)
				-- then the transform time count down increases, and any started
				-- re-transformation process fades away
				if cur_target_distance <= COPY_ENABLED_DISTANCE or (closest_p and closest_p.mo.skin == ds.copied_skin) then
					ds.transform_time = min(BASE_TRANSFORM_TIME, $ + 1)
					ds.transform_progress = max(0, $ - 1)
					continue
				end

				-- if the transfromed ditto's previously copied traget is still in range, increase the
				--  transfrom time countdown, otherwise decrease.
				if ds.istransformed then
					if cur_target_distance <= COPY_ENABLED_DISTANCE then
						ds.transform_time = min(BASE_TRANSFORM_TIME, $ + 1)
					elseif cur_target_distance > RETAIN_MEMORY_DISTANCE then
						ds.transform_time = max(0, $ - 1)
						ds.untransforming = true
					end

					-- oh yeah, and there's a minimum amount of time a ditto stays transformed,
					--  so if we haven't reached that, abort anything else
					if ds.real_transform_time < MIN_TRANSFORM_TIME then continue end
				end

				-- if there's a new target in range (closest), we update the transforming gauge
				-- and if there's not we just decrease is it until it reaches 0
				if ds.copying_target and ds.copying_target.mo.skin ~= ds.copied_skin and (closest_p and ds.copying_target.mo.skin == closest_p.mo.skin) then
					ds.transform_progress = $ + 1
				else
					if closest_p then
						ds.transform_progress = 0
						ds.copying_target = closest_p
					else
						ds.transform_progress = max(0, $ - 1)
						if ds.transform_progress <= 0 then
							ds.copying_target = nil
						end
					end
				end

				-- if the transform count down reaches 0, the players detransforms
				if ds.transform_time == 0 then
					playerDetransforms(player, true)
					continue
				end

				-- if the transforming gauge is full, leave our old self behind and
				-- embrace our new, not even final form
				if ds.transform_progress >= TRANFORMING_DURATION then
					playerCopies(player, ds.copying_target)
				end
			end
		end
	end

	_cleaning = 1
end)

addHook("IntermissionThinker", function()
	if doconfig then sendconfig(); doconfig = false end
	cleaning(2)
end)

addHook("VoteThinker", function()
	if doconfig then sendconfig(); doconfig = false end
	cleaning(3)
end)

-- we'll put the sprites of the HUD transfo indicators in here
local a_copyIndicator
local STAT_DISPLAY_DURATION = TICRATE+(TICRATE/3)

-- let's draw the stuff on the hud when we need it!
hud.add(function(v, p, c)
	-- oh right! make sure we have all the sprite patch we need before anything else
	if not a_copyIndicator then
		a_copyIndicator = {}
		for i = 0, 5 do
			table.insert(a_copyIndicator, v.cachePatch("DITPRG"..i))
		end
	end

	local ds = p.dittostuff
	if not (ds and p.mo) or finished(p) or ds.transform_locked then return end

	local copyProgress = ds.transform_progress
	local transform_time = ds.transform_time
	local real_time = ds.real_transform_time

	--[[ this was for debug only
	local s = ""
	if transform_time then
		s = "transfTime: "..transform_time
		.." ("..tostring(ds.copied_target and ds.copied_target.name)..")"
		v.drawString(160-v.stringWidth(s, nil, "small")/2, 10, s, V_ALLOWLOWERCASE, "small")
	end
	if real_time then
		s = (real_time < MIN_TRANSFORM_TIME and "\x85" or "").."realTime: "..real_time
		v.drawString(160-v.stringWidth(s, nil, "small")/2, 15, s, V_ALLOWLOWERCASE, "small")
	end
	if copyProgress then
		s = "transfProgress: "..copyProgress
		.." ("..tostring(ds.copying_target and ds.copying_target.name)..")"
		v.drawString(160-v.stringWidth(s, nil, "small")/2, 20, s, V_ALLOWLOWERCASE, "small")
	end
	--]]

	-- okay, let's draw the ditto HUD stuff now
	-- drawing the transformation cues on the HUD, y'a know?
	local basex, basey = SG_GetScreenCoords(v, p, c, p.mo, p.mo.height)
	if not basex then return end

	local vflags = V_NOSCALESTART|V_OFFSET
	basex, basey = $>>FRACBITS, $>>FRACBITS

	-- now with the annoying stuff out of the ways,
	-- let's draw the 'transforming progress' sprite, if needed
	if copyProgress then
		local pstep = min(#a_copyIndicator, copyProgress)
		local t, colormap = ds.copying_target
		if t and t.valid and t.mo then
			colormap = v.getColormap(t.mo.skin, t.skincolor)
		end

		v.draw(basex, basey, a_copyIndicator[pstep], vflags, colormap)
	elseif transform_time then
		-- or maybe draw the 'detransforming soon' cue?
		if transform_time < TRANFORMING_DURATION and ds.untransforming then
			local str = ""
			local pstep = min(3, transform_time/(TRANFORMING_DURATION/3)+1)
			for i = pstep, 3 do
				str = $.."."
			end
			v.drawString(basex - v.stringWidth(str, vflags), basey - 20/v.dupy(), str, vflags)
		end

		-- also if we'd just transformed give a cue to show the new stats
		--  (let's not be assholes about it hehe)
		if real_time and real_time < STAT_DISPLAY_DURATION and leveltime >= GO_TIME+TICRATE then
			local str_speed = "\136"..p.kartspeed
			local str_weight = "\135"..p.kartweight

			local d = v.dupx()*real_time/3 + 1
			local xs, ys = basex - (3+d), basey - d
			local xw, yw = basex + (3+d), basey + d + 6

			local trans = V_40TRANS + (4*real_time/STAT_DISPLAY_DURATION)*V_10TRANS

			v.drawString(xs - 4, ys, str_speed, trans | vflags)
			v.drawString(xw - 4, yw, str_weight, trans | vflags)
		end
	end
end)

-- meh
addHook("NetVars", function(net)
	_cleaning = net($)
end)
