/* 
Script for Luna-P V2.0
	by Steel Titanium, adapted by Bartman3010
	
Creates a Luna-P object to follow Chibi-Usa and react to the character in-game.
*/

//Freeslot the required stuff
freeslot(
	"SPR_LUNP",
	"S_LUNAPIDLE",
	"S_LUNAPFLY",
	"S_LUNAPHAPPY1",
	"S_LUNAPHAPPY2",
	"S_LUNAPHAPPY3",
	"S_LUNAPHAPPY4",
	"S_LUNAPSAD1",
	"S_LUNAPSAD2",
	"S_LUNAPSAD3",
	"S_LUNAPSAD4",
	"MT_LUNAP"
)

mobjinfo[MT_LUNAP] = {
	doomednum = -1,
	spawnstate = S_LUNAPIDLE,
	spawnhealth = 1,
	painchance = 200,
	speed = 3,
	radius = 16*FRACUNIT,
	height = 32*FRACUNIT,
	dispoffset = 0,
	mass = 100,
	damage = 0,
	reactiontime = 32,
	activesound = sfx_None,
	flags = MF_NOGRAVITY|MF_FLOAT|MF_SCENERY|MF_NOCLIPHEIGHT|MF_DONTENCOREMAP,
	raisestate = S_NULL
}

states[S_LUNAPIDLE] = {
	sprite = SPR_LUNP,
	frame = A,
	tics = -1,
	var1 = 1,
	var2 = 2,
	nextstate = S_LUNAPIDLE
}

states[S_LUNAPFLY] = {
	sprite = SPR_LUNP,
	frame = A,
	tics = -1,
	var1 = 1,
	var2 = 2,
	nextstate = S_LUNAPFLY
}

states[S_LUNAPSAD1] = {
	sprite = SPR_LUNP,
	frame = F,
	tics = 3,
	var1 = 0,
	var2 = 0,
	nextstate = S_LUNAPSAD2
}

states[S_LUNAPSAD2] = {
	sprite = SPR_LUNP,
	frame = G,
	tics = 3,
	var1 = 0,
	var2 = 0,
	nextstate = S_LUNAPSAD3
}

states[S_LUNAPSAD3] = {
	sprite = SPR_LUNP,
	frame = H,
	tics = 3,
	var1 = 0,
	var2 = 0,
	nextstate = S_LUNAPSAD4
}

states[S_LUNAPSAD4] = {
	sprite = SPR_LUNP,
	frame = G,
	tics = 3,
	var1 = 0,
	var2 = 0,
	nextstate = S_LUNAPSAD1
}

states[S_LUNAPHAPPY1] = {
	sprite = SPR_LUNP,
	frame = B,
	tics = 4,
	nextstate = S_LUNAPHAPPY2
}

states[S_LUNAPHAPPY2] = {
	sprite = SPR_LUNP,
	frame = C,
	tics = 4,
	nextstate = S_LUNAPHAPPY3
}

states[S_LUNAPHAPPY3] = {
	sprite = SPR_LUNP,
	frame = D,
	tics = 4,
	nextstate = S_LUNAPHAPPY4
}

states[S_LUNAPHAPPY4] = {
	sprite = SPR_LUNP,
	frame = E,
	tics = 4,
	nextstate = S_LUNAPHAPPY1
}

					   
				   
		   
		  
						
 

addHook("MobjThinker", function(mo)
	if mo and mo.skin == "chibiusa" and mo.player //Skin check
		local player = mo.player
		if not player.mo.lunap
			player.mo.lunap = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_LUNAP)
			player.mo.lunap.angle = player.mo.angle
			player.mo.lunap.flags2 = player.mo.flags2
			player.mo.lunap.destscale = player.mo.destscale
			player.mo.lunap.target = player.mo
		end
	elseif mo and mo.skin ~= "chibiusa" and mo.lunap and mo.lunap.valid then //Not using Chibi-Usa but still has a Luna-P with them.
		P_RemoveMobj(mo.lunap)
		mo.lunap = nil
	end
end, MT_PLAYER)

addHook("MobjThinker", function(mo)
	-- Thinker for Luna-P
	if mo and mo.valid then
		if mo.target and mo.target.valid and mo.target.player and mo.target.player.valid then
			local tg = mo.target
			local totz

			-- Colorize the Luna-P if the player has invincibility, grow etc.
			mo.color = tg.color
			mo.colorized = tg.colorized

			-- Check for gravity flip
			if (mo.eflags & MFE_VERTICALFLIP)
				totz = (tg.z) - (tg.height)
			else
				totz = (tg.z) + (tg.height)
			end
			-- Luna-P's rotation position when player is idle
			local mangle = FixedAngle(((leveltime * 5) % 360) << FRACBITS)
			mo.destscale = tg.destscale
			mo.angle = tg.angle
			mo.eflags = tg.eflags
			mo.flags2 = tg.flags2

			if (tg.player.exiting > 0 and K_IsPlayerLosing(tg.player) == true) or tg.player.kartstuff[k_spinouttimer] > 0 or tg.health == 0
				if not (mo.state >= S_LUNAPSAD1 and mo.state <= S_LUNAPSAD4)
					mo.state = S_LUNAPSAD1
				end
				mo.z = totz
				P_MoveOrigin(mo, tg.x, tg.y, totz)
				mo.angle = tg.angle
			end

			if tg.health > 0 and tg.player.kartstuff[k_spinouttimer] == 0 then
				if tg.player.speed > 0 or (tg.momx + tg.momy + tg.momz) > 0 then // If player is moving
					if mo.prevtpos == nil then
						mo.prevtpos = {}
					end

					if not (mo.state == S_LUNAPFLY)
						mo.state = S_LUNAPFLY
					end
					local trailby = 2

					if #mo.prevtpos >= trailby then
						local e = mo.prevtpos[trailby]
						local etotz

						if (mo.eflags & MFE_VERTICALFLIP)
							etotz = e.z - (tg.height) - (sin(FixedAngle(leveltime*5*FRACUNIT))*8)
						else
							etotz = e.z + (tg.height) + (sin(FixedAngle(leveltime*5*FRACUNIT))*8)
						end

						mo.prevtpos[trailby] = nil
						P_MoveOrigin(mo, e.x, e.y, etotz)
					end

					table.insert(mo.prevtpos, 1, {x = tg.x, y = tg.y, z = tg.z})
				else -- If player is idle
					P_MoveOrigin(mo, tg.x, tg.y, totz)

					if tg.player.exiting > 0 and K_IsPlayerLosing(tg.player) == false
						if not (mo.state >= S_LUNAPHAPPY1 and mo.state <= S_LUNAPHAPPY4)
							mo.state = S_LUNAPHAPPY1
						end
					elseif tg.player.exiting > 0 and K_IsPlayerLosing(tg.player) == true
						if not (mo.state >= S_LUNAPSAD1 and mo.state <= S_LUNAPSAD4)
							mo.state = S_LUNAPSAD1
						end
					else
						if not (mo.state == S_LUNAPIDLE)
							mo.state = S_LUNAPIDLE
						end
					end

					mo.angle = tg.angle
					mo.momx = sin(mangle) * (tg.radius>>FRACBITS)
					mo.momy = cos(mangle) * (tg.radius>>FRACBITS)
				end
			end
		end
	end

	if mo and (not mo.target or not mo.target.valid) or (not mo.target.player or not mo.target.valid) then
		P_RemoveMobj(mo)
    end

end, MT_LUNAP)